#include "iMS_EnhancedSequences.h"
#include "Supervisors.h"
#include "lvgl.h"

#include <sstream>
#include <random>

extern "C" lv_font_t freeserif_16;

#define MAX_FONT_GLYPHS 8086

static ImageGroup grp;

static bool get_image_from_font_map(const lv_font_t* font, Image& img_out, unsigned int unicode_index) {
	lv_coord_t col, row;
	uint8_t bitmask_init;
	uint8_t bitmask;
	uint8_t letter_px;

	img_out.clear();

	lv_font_glyph_dsc_t dsc;
	bool g_ret = lv_font_get_glyph_dsc(font, &dsc, unicode_index, 0);
	if (g_ret == false) return false;

	const uint8_t* map_p = lv_font_get_glyph_bitmap(font, unicode_index);
	if (map_p == NULL) return false;

	switch (dsc.bpp) {
	case 1:
		bitmask_init = 0x80;
		break;
	case 2:
		bitmask_init = 0xC0;
		break;
	case 4:
		bitmask_init = 0xF0;
		break;
	case 8:
		bitmask_init = 0xFF;
		break;
	default: return false; /*Invalid bpp. Can't render the letter*/
	}

	MHz fx, fx_init = freq_centre_X - (dsc.box_w / 2) * freq_interval_X;
	MHz fy, fy_init = freq_centre_Y + (dsc.box_h / 2) * freq_interval_Y;
	FAP fapx, fapy;

	uint16_t col_bit = 0;
	for (row = 0, fy = fy_init; row < dsc.box_h; row++, fy = fy - freq_interval_Y) {
		bitmask = bitmask_init;

		for (col = 0, fx = fx_init; col < dsc.box_w; col++, fx = fx + freq_interval_X) {
			letter_px = (*map_p & bitmask) >> (8 - col_bit - dsc.bpp);

			// Create frequency from row/col position
			fapx.freq = fx;
			fapy.freq = fy;

			// Create amplitude from pixel data
			fapx.ampl = fapy.ampl = Percent(20.0 + 80.0 * sqrt((double)letter_px / (1 << dsc.bpp)));

			img_out.AddPoint(ImagePoint(fapx, fapx, fapy, fapy));

			if (col_bit < 8 - dsc.bpp) {
				col_bit += dsc.bpp;
				bitmask = bitmask >> dsc.bpp;
			}
			else {
				col_bit = 0;
				bitmask = bitmask_init;
				map_p++;
			}
		}
	}

	return true;
}

static bool wait_for_image_download(IMSSystem& ims, const Image& img)
{
	ImageDownload* dl = new ImageDownload(ims, img);
	ImageDownloadSupervisor ds;

	dl->ImageDownloadEventSubscribe(ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
	dl->ImageDownloadEventSubscribe(ImageDownloadEvents::DOWNLOAD_ERROR, &ds);

	if (!dl->StartDownload()) {
		std::cout << "Error starting download" << std::endl;
		dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
		dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_ERROR, &ds);
		delete dl;
		return false;
	}

	int timeout = 1000;
	while (ds.Busy()) {
		std::this_thread::sleep_for(std::chrono::milliseconds(5));
		if (timeout-- < 0) {
			std::cout << "Timeout on Image Download." << std::endl;
			return false;
		}
	}

	dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
	dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_ERROR, &ds);

	if (ds.Error()) return false;

	delete dl;
	return true;
}

void createFontMapImageGroup()
{
	//
	// Create Image group with data from Unicode character map
	//
	int n = number_of_images_in_group;
	if (n > MAX_FONT_GLYPHS) n = MAX_FONT_GLYPHS;

	uint32_t unicode_index = 0;
	std::cout << "Creating Image Group (" << n << " images) " << std::flush;
	do {
		std::shared_ptr<Image> img = std::make_shared<Image>();
		if (get_image_from_font_map(&freeserif_16, *img, unicode_index)) {
			if (img->Size() < min_image_size) {
				img->Clear();
			}
			else {
				std::stringstream ss;
				ss << "U+" << std::hex << unicode_index;
				img->Name() = ss.str();
				img->ClockRate(image_clock_rate);
				grp.AddImage(*img);
			}
		}
		unicode_index++;
	} while (grp.Size() < n);
}

void downloadFontMapToIMS(std::shared_ptr<iMS::IMSSystem> myiMS)
{
	int i = 0;
	std::cout << "Downloading" << std::flush;
	for (const auto& img : grp) {
		if (!wait_for_image_download(*myiMS, img)) {
			std::cout << "Aborting.. <press any key>" << std::endl;
			char c;
			std::cin >> c;
			std::exit(-1);
		}

		if (!(i++ % 25)) std::cout << "." << std::flush;
	}
}

iMS::Image& getImage(int index)
{
	static std::default_random_engine rand_gen;
	std::uniform_int_distribution<int> rand_int(0, grp.size() - 1);

	if ((index < 0) || (index > grp.size())) {
		return grp[rand_int(rand_gen)];
	} else {
		return grp[index];
	}
}

const iMS::Image& getImage(std::array<uint8_t, 16> uuid) {
	auto it = std::find_if(std::begin(grp), std::end(grp), [&](const Image& img) { return img.GetUUID() == uuid; });
	return *it;
}